==> policyengine_us-0.687.0-py3-none-any.whl: policyengine_us/data/datasets/cps/enhanced_cps/calibrate.py <==
import torch
import pandas as pd
import numpy as np
from policyengine_us import Microsimulation
import plotly.express as px
from tqdm import tqdm
from policyengine_core.data import Dataset
from policyengine_us.data import CPS_2023
import numpy as np
from policyengine_us import Microsimulation
import pandas as pd
from .process_puf import (
    FINANCIAL_SUBSET as FINANCIAL_VARIABLES,
    puf_imputed_cps_person_level,
)
from pathlib import Path
from typing import Tuple


def generate_model_variables(
    dataset: str, time_period: str = "2023", no_weight_adjustment: bool = False
) -> Tuple:
    """Generates variables needed for the calibration process.

    Args:
        dataset (str): The name of the dataset to use.
        time_period (str, optional): The time period to use. Defaults to "2023".
        no_weight_adjustment (bool, optional): Whether to skip the weight adjustment. Defaults to False.

    Returns:
        household_weights (torch.Tensor): The household weights.
        weight_adjustment (torch.Tensor): The weight adjustment.
        values_df (pd.DataFrame): A 2D array of values to transform household weights into statistical predictions.
        targets (dict): A dictionary of names and target values for the statistical predictions.
        targets_array (dict): A 1D array of target values for the statistical predictions.
        equivalisation_factors_array (dict): A 1D array of equivalisation factors for the statistical predictions to normalise the targets.
    """
    simulation = Microsimulation(dataset=dataset)
    simulation.default_calculation_period = time_period
    parameters = simulation.tax_benefit_system.parameters.calibration(
        f"{time_period}-01-01"
    )

    household_weights = torch.tensor(
        simulation.calculate("household_weight").values, dtype=torch.float32
    )
    weight_adjustment = torch.tensor(
        np.random.random(household_weights.shape) * 10,
        requires_grad=True,
        dtype=torch.float32,
    )

    if no_weight_adjustment:
        weight_adjustment = torch.zeros_like(household_weights)

    values_df = pd.DataFrame()
    targets = {}
    equivalisation = {}

    # We need to normalise the targets. Common regression targets are often 1e1 to 1e3 (this informs the scale of the learning rate).
    COUNT_HOUSEHOLDS = household_weights.sum().item()
    FINANCIAL_EQUIVALISATION = COUNT_HOUSEHOLDS
    POPULATION_EQUIVALISATION = COUNT_HOUSEHOLDS / 1e5

    is_filer = simulation.calculate("tax_unit_is_filer").values
    household_has_filers = (
        simulation.map_result(is_filer, "tax_unit", "household") > 0
    )

    for variable_name in FINANCIAL_VARIABLES:
        if variable_name not in parameters.gov.irs.soi:
            continue
        label = (
            simulation.tax_benefit_system.variables[variable_name].label
            + " (IRS SOI)"
        )
        values_df[label] = (
            simulation.calculate(variable_name, map_to="household").values
            * household_has_filers
        )
        targets[label] = parameters.gov.irs.soi[variable_name]
        equivalisation[label] = FINANCIAL_EQUIVALISATION

    # Program spending from CBO baseline projections

    PROGRAMS = [
        "income_tax",
        "snap",
        "social_security",
        "ssi",
        "unemployment_compensation",
    ]

    for variable_name in PROGRAMS:
        label = (
            simulation.tax_benefit_system.variables[variable_name].label
            + " (CBO)"
        )
        values_df[label] = simulation.calculate(
            variable_name, map_to="household"
        ).values
        targets[label] = parameters.gov.cbo[variable_name]
        equivalisation[label] = FINANCIAL_EQUIVALISATION

    snap_participation = parameters.gov.usda.snap.participation
    ssi_participation = parameters.gov.ssa.ssi.participation
    ss_participation = parameters.gov.ssa.social_security.participation

    for program, participation in zip(
        ["snap", "ssi", "social_security"],
        [snap_participation, ssi_participation, ss_participation],
    ):
        label = simulation.tax_benefit_system.variables[program].label
        entity_level = simulation.tax_benefit_system.variables[
            program
        ].entity.key
        entity_level_value = simulation.calculate(program)
        values_df[f"{label} participants"] = simulation.map_result(
            entity_level_value > 0, entity_level, "household"
        )
        targets[f"{label} participants"] = participation
        equivalisation[f"{label} participants"] = POPULATION_EQUIVALISATION

    demographics_sim = Microsimulation(dataset="cps_2023")

    # Total population
    values_df["U.S. population"] = simulation.calculate(
        "people", map_to="household"
    ).values
    targets["U.S. population"] = parameters.gov.census.populations.total
    equivalisation["U.S. population"] = POPULATION_EQUIVALISATION

    # Population by 10-year age group and sex
    age_cps = demographics_sim.calculate("age").values
    is_male_cps = demographics_sim.calculate("is_male")
    age = simulation.calculate("age").values
    is_male = simulation.calculate("is_male")
    population_in_21 = demographics_sim.tax_benefit_system.parameters.calibration.gov.census.populations.total(
        "2021-01-01"
    )
    population_growth_since_21 = (
        parameters.gov.census.populations.total / population_in_21
    )
    cps_household_weights = demographics_sim.calculate(
        "household_weight"
    ).values
    for lower_age_group in range(0, 90, 10):
        for possible_is_male in (True, False):
            in_age_range = (age >= lower_age_group) & (
                age < lower_age_group + 5
            )
            in_sex_category = is_male == possible_is_male
            count_people_in_range = simulation.map_result(
                in_age_range * in_sex_category, "person", "household"
            )
            in_age_range_cps = (age_cps >= lower_age_group) & (
                age_cps < lower_age_group + 5
            )
            in_sex_category_cps = is_male_cps == possible_is_male
            count_people_in_range_cps = demographics_sim.map_result(
                in_age_range_cps * in_sex_category_cps, "person", "household"
            )
            count_people_in_range = simulation.map_result(
                in_age_range * in_sex_category, "person", "household"
            )
            sex_category = "male" if possible_is_male else "female"
            name = f"{lower_age_group} to {lower_age_group + 5} and {sex_category} population"
            values_df[name] = count_people_in_range
            targets[name] = (
                cps_household_weights * count_people_in_range_cps
            ).sum() * population_growth_since_21
            equivalisation[name] = POPULATION_EQUIVALISATION

    # Household population by number of adults and children

    household_count_adults_cps = demographics_sim.map_result(
        age_cps >= 18, "person", "household"
    )
    household_count_children_cps = demographics_sim.map_result(
        age_cps < 18, "person", "household"
    )

    household_count_adults = simulation.map_result(
        age >= 18, "person", "household"
    )
    household_count_children = simulation.map_result(
        age < 18, "person", "household"
    )

    for count_adults in range(1, 3):
        for count_children in range(0, 4):
            in_criteria = (
                (household_count_adults == count_adults)
                * (household_count_children == count_children)
                * 1.0
            )
            in_criteria_cps = (
                (household_count_adults_cps == count_adults)
                * (household_count_children_cps == count_children)
                * 1.0
            )
            name = f"{count_adults}-adult, {count_children}-child household population"
            values_df[name] = in_criteria
            targets[name] = (
                cps_household_weights * in_criteria_cps
            ).sum() * population_growth_since_21
            equivalisation[name] = POPULATION_EQUIVALISATION

    # Number of tax returns by AGI category

    # is_filer

    agi = simulation.calculate("adjusted_gross_income").values

    BOUNDS = [
        -np.inf,
        1,
        5e3,
        10e3,
        15e3,
        20e3,
        25e3,
        30e3,
        40e3,
        50e3,
        75e3,
        100e3,
        200e3,
        500e3,
        1e6,
        1.5e6,
        2e6,
        5e6,
        10e6,
        np.inf,
    ]
    COUNTS = [
        4_098_522,
        8_487_025,
        8_944_908,
        10_056_377,
        9_786_580,
        8_863_570,
        8_787_576,
        16_123_068,
        12_782_334,
        22_653_934,
        14_657_726,
        24_044_481,
        9_045_567,
        1_617_144,
        376_859,
        156_020,
        233_838,
        63_406,
        45_404,
    ]
    VALUES = [
        -171_836_364,
        19_987_243,
        67_651_359,
        125_912_056,
        170_836_129,
        199_508_960,
        241_347_179,
        561_386_434,
        573_155_378,
        1_392_395_599,
        1_271_699_391,
        3_297_058_075,
        2_619_188_471,
        1_092_599_034,
        454_552_875,
        268_278_123,
        698_923_219,
        435_242_550,
        1_477_728_359,
        13_879_929_368,
        -12_835_378,
        451_204,
        1_358_544,
        14_362_205,
        57_643_020,
        101_727_915,
        141_934_070,
        382_385_416,
        457_336_377,
        1_238_178_360,
        1_206_614_503,
        3_252_746_502,
        2_613_795_014,
        1_091_571_914,
        3_332_659_702,
    ]

    for i in range(len(BOUNDS) - 1):
        lower_bound = BOUNDS[i]
        upper_bound = BOUNDS[i + 1]
        in_range = (agi >= lower_bound) * (agi < upper_bound) * is_filer
        household_filers = simulation.map_result(
            in_range, "tax_unit", "household"
        )
        if lower_bound == -np.inf:
            lower_bound_str = "negative infinity"
        else:
            lower_bound_str = f"{lower_bound:,.0f}"
        name = f"tax returns with AGI between ${lower_bound_str} and ${upper_bound:,.0f}"
        values_df[name] = household_filers
        targets[name] = COUNTS[i] * population_growth_since_21
        equivalisation[name] = POPULATION_EQUIVALISATION

        agi_in_range = agi * in_range
        household_agi = simulation.map_result(
            agi_in_range, "tax_unit", "household"
        )
        name = f"total AGI from tax returns with AGI between ${lower_bound_str} and ${upper_bound:,.0f}"
        values_df[name] = household_agi
        targets[name] = VALUES[i] * population_growth_since_21 * 1e3
        equivalisation[name] = FINANCIAL_EQUIVALISATION

    targets_array = torch.tensor(list(targets.values()), dtype=torch.float32)
    equivalisation_factors_array = torch.tensor(
        list(equivalisation.values()), dtype=torch.float32
    )

    return (
        household_weights,
        weight_adjustment,
        values_df,
        targets,
        targets_array,
        equivalisation_factors_array,
    )


def aggregate(
    adjusted_weights: torch.Tensor, values: pd.DataFrame
) -> torch.Tensor:
    broadcasted_weights = adjusted_weights.reshape(-1, 1)
    weighted_values = torch.matmul(
        broadcasted_weights.T, torch.tensor(values.values, dtype=torch.float32)
    )
    return weighted_values


def get_snapshot(
    dataset: str,
    time_period: str = "2023",
) -> pd.DataFrame:
    """Returns a snapshot of the training metrics without training the model.

    Args:
        dataset (str): The name of the dataset to use.
        time_period (str, optional): The time period to use. Defaults to "2023".

    Returns:
        pd.DataFrame: A DataFrame containing the training metrics.
    """
    (
        household_weights,
        weight_adjustment,
        values_df,
        targets,
        targets_array,
        equivalisation_factors_array,
    ) = generate_model_variables(
        dataset, time_period, no_weight_adjustment=True
    )
    adjusted_weights = torch.relu(household_weights + weight_adjustment)
    result = (
        aggregate(adjusted_weights, values_df) / equivalisation_factors_array
    )
    target = targets_array / equivalisation_factors_array
    current_aggregates = (
        (result * equivalisation_factors_array).detach().numpy()[0]
    )
    loss = torch.mean(((result / target - 1) ** 2) * np.log2(np.abs(target)))
    current_loss = loss.item()
    return pd.DataFrame(
        {
            "name": list(targets.keys()) + ["total"],
            "value": list(current_aggregates) + [current_loss],
            "target": list(targets.values()) + [0],
            "time_period": time_period,
        }
    )


def calibrate(
    dataset: str,
    time_period: str = "2023",
    training_log_path: str = "training_log.csv.gz",
    learning_rate: float = 2e1,
    epochs: int = 20_000,
) -> np.ndarray:
    (
        household_weights,
        weight_adjustment,
        values_df,
        targets,
        targets_array,
        equivalisation_factors_array,
    ) = generate_model_variables(dataset, time_period)
    training_log_path = Path(training_log_path)
    if training_log_path.exists():
        training_log_df = pd.read_csv(training_log_path, compression="gzip")
    else:
        training_log_df = pd.DataFrame()

    progress_bar = tqdm(range(epochs), desc="Calibrating weights")
    optimizer = torch.optim.Adam([weight_adjustment], lr=learning_rate)
    for i in progress_bar:
        adjusted_weights = torch.relu(household_weights + weight_adjustment)
        result = (
            aggregate(adjusted_weights, values_df)
            / equivalisation_factors_array
        )
        target = targets_array / equivalisation_factors_array
        loss = torch.mean(
            ((result / target - 1) ** 2) * np.log2(np.abs(target))
        )
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        if i % 10 == 0:
            current_loss = loss.item()
            progress_bar.set_description_str(
                f"Calibrating weights | Loss = {current_loss:,.3f}"
            )
        if i % 2_000 == 0:
            current_aggregates = (
                (result * equivalisation_factors_array).detach().numpy()[0]
            )
            training_log_df = pd.concat(
                [
                    training_log_df,
                    pd.DataFrame(
                        {
                            "name": list(targets.keys()) + ["total"],
                            "epoch": [i] * len(targets) + [i],
                            "value": list(current_aggregates) + [current_loss],
                            "target": list(targets.values()) + [0],
                            "time_period": time_period,
                        }
                    ),
                ]
            )
            training_log_df.to_csv(training_log_path, compression="gzip")

    return adjusted_weights.detach().numpy()

==> policyengine_us-0.708.5-py3-none-any.whl: policyengine_us/data/datasets/cps/enhanced_cps/calibrate.py <==
import torch
import pandas as pd
import numpy as np
from policyengine_us import Microsimulation
import plotly.express as px
from tqdm import tqdm
from policyengine_core.data import Dataset
from policyengine_us.data import CPS_2023
import numpy as np
from policyengine_us import Microsimulation
import pandas as pd
from .process_puf import (
    FINANCIAL_SUBSET as FINANCIAL_VARIABLES,
    puf_imputed_cps_person_level,
)
from pathlib import Path
from typing import Tuple
from policyengine_us.data.datasets.cps.enhanced_cps.loss import (
    generate_model_variables,
)


def aggregate(
    adjusted_weights: torch.Tensor, values: pd.DataFrame
) -> torch.Tensor:
    broadcasted_weights = adjusted_weights.reshape(-1, 1)
    weighted_values = torch.matmul(
        broadcasted_weights.T, torch.tensor(values.values, dtype=torch.float32)
    )
    return weighted_values


def calibrate(
    dataset: str,
    time_period: str = "2023",
    training_log_path: str = "training_log.csv.gz",
    learning_rate: float = 2e1,
    epochs: int = 20_000,
) -> np.ndarray:
    (
        household_weights,
        weight_adjustment,
        values_df,
        targets,
        targets_array,
        equivalisation_factors_array,
    ) = generate_model_variables(dataset, time_period)
    household_weights = torch.tensor(household_weights, dtype=torch.float32)
    weight_adjustment = torch.tensor(
        weight_adjustment, dtype=torch.float32, requires_grad=True
    )
    targets_array = torch.tensor(targets_array, dtype=torch.float32)
    equivalisation_factors_array = torch.tensor(
        equivalisation_factors_array, dtype=torch.float32
    )
    training_log_path = Path(training_log_path)
    if training_log_path.exists():
        training_log_df = pd.read_csv(training_log_path, compression="gzip")
    else:
        training_log_df = pd.DataFrame()

    progress_bar = tqdm(range(epochs), desc="Calibrating weights")
    optimizer = torch.optim.Adam([weight_adjustment], lr=learning_rate)
    for i in progress_bar:
        adjusted_weights = torch.relu(household_weights + weight_adjustment)
        result = (
            aggregate(adjusted_weights, values_df)
            / equivalisation_factors_array
        )
        target = targets_array / equivalisation_factors_array
        loss = torch.mean(
            ((result / target - 1) ** 2) * np.log2(np.abs(target))
        )
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        if i % 10 == 0:
            current_loss = loss.item()
            progress_bar.set_description_str(
                f"Calibrating weights | Loss = {current_loss:,.3f}"
            )
        if i % 2_000 == 0:
            current_aggregates = (
                (result * equivalisation_factors_array).detach().numpy()[0]
            )
            training_log_df = pd.concat(
                [
                    training_log_df,
                    pd.DataFrame(
                        {
                            "name": list(targets.keys()) + ["total"],
                            "epoch": [i] * len(targets) + [i],
                            "value": list(current_aggregates) + [current_loss],
                            "target": list(targets.values()) + [0],
                            "time_period": time_period,
                        }
                    ),
                ]
            )
            training_log_df.to_csv(training_log_path, compression="gzip")

    return adjusted_weights.detach().numpy()
